;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "ghostel" "20260505.1350"
  "Terminal emulator powered by libghostty."
  '((emacs "28.1"))
  :url "https://github.com/dakra/ghostel"
  :commit "cfd7fe7fccf01d0babc53811def823fb52420f40"
  :revdesc "cfd7fe7fccf0"
  :keywords '("terminals")
  :authors '(("Daniel Kraus" . "daniel@kraus.my"))
  :maintainers '(("Daniel Kraus" . "daniel@kraus.my")))
